# This file is part of thymiodirect.
# Copyright 2020 ECOLE POLYTECHNIQUE FEDERALE DE LAUSANNE,
# Miniature Mobile Robots group, Switzerland
# Author: Yves Piguet
#
# SPDX-License-Identifier: BSD-3-Clause

"""
Communication with Thymio II robot
==================================

This module provides support to connect to a Thymio II robot with its native
binary protocol via a serial port (virtual port over a wired USB or wireless
USB dongle) or a TCP port (asebaswitch or Thymio simulator).

Example
-------

# import the required classes
from thymiodirect import Connection
from thymiodirect import Thymio

# set the serial port the Thymio is connected to
# (depending on your configuration, the default port is not what you want)
port = Connection.serial_default_port()

# create a Thymio connection object with a callback to be notified when
# the robot is ready and start the connection (or just wait a few seconds)
th = Thymio(serial_port=port,
            on_connect=lambda node_id:print(f"{node_id} is connected"))
th.connect()

# get id of the first (or only) Thymio
id = th.first_node()

# get a variable
th[id]["prox.horizontal"]

# set a variable (scalar or array)
th[id]["leds.top"] = [0, 0, 32]

# define a function called after new variable values have been fetched
prox_prev = 0
def obs(node_id):
    global prox_prev
    prox = (th[node_id]["prox.horizontal"][5]
            - th[node_id]["prox.horizontal"][2]) // 10
    if prox != prox_prev:
        th[node_id]["motor.left.target"] = prox
        th[node_id]["motor.right.target"] = prox
        print(prox)
        if prox > 5:
            th[id]["leds.top"] = [0, 32, 0]
        elif prox < -5:
            th[id]["leds.top"] = [32, 32, 0]
        elif abs(prox) < 3:
            th[id]["leds.top"] = [0, 0, 32]
        prox_prev = prox
    if th[node_id]["button.center"]:
        print("button.center")
        os._exit(0) # forced exit despite coroutines

# install this function
th.set_variable_observer(id, obs)
"""

from thymiodirect.connection import Connection
from thymiodirect.thymio import Thymio



# This file is part of thymiodirect.
# Copyright 2020 ECOLE POLYTECHNIQUE FEDERALE DE LAUSANNE,
# Miniature Mobile Robots group, Switzerland
# Author: Yves Piguet
#
# SPDX-License-Identifier: BSD-3-Clause

"""
Communication with Thymio via serial port or tcp
Author: Yves Piguet, EPFL
"""

import asyncio
import threading
import time
from typing import List

from thymiodirect.connection import Connection
from thymiodirect.assembler import Assembler


class Thymio:
    """
    Thymio is a helper object for communicating with one or several Thymios
    connected to a single port.
    """

    class _ThymioProxy:
        """
        _ThymioProxy is a proxy to a Thymio object to call user-defined
        functions from asyncio coroutines.
        """

        def __init__(self, thymio: "thymio.Thymio"):
            """
            Construct a new __ThymioProxy object.

            :param thymio: Thymio object
            """

            self.thymio = thymio
            self.connection = None
            self.loop = asyncio.get_event_loop()
            self.nodes = set()

        def run(self):
            """
            Run the asyncio loop forever, taking care of executing callbacks
            registered for the Thymio connection.
            """

            async def on_connection_changed(node_id, connected: bool):
                """
                Do what's required when the Thymio connection changes.
                """
                if connected:
                    self.nodes.add(node_id)
                    if self.thymio.on_connect_cb:
                        self.thymio.on_connect_cb(node_id)
                else:
                    self.nodes.remove(node_id)
                    if self.thymio.on_disconnect_cb:
                        self.thymio.on_disconnect_cb(node_id)

            async def on_variables_received(node_id: int):
                """
                Do what's required when new variables have been received.
                """
                if node_id in self.thymio.variable_observers:
                    variable_observer = self.thymio.variable_observers[node_id]
                    variable_observer(node_id)

            async def on_user_event(node_id: int, event_id: int, event_args: List[int]):
                """
                Do what's required when an event has been received from the Thymio.
                """
                if node_id in self.thymio.user_event_listeners:
                    user_event_listener = self.thymio.user_event_listeners[node_id]
                    user_event_listener(node_id, event_id, event_args)

            if self.thymio.use_tcp:
                self.connection = Connection.tcp(host=self.thymio.host,
                                                 port=self.thymio.tcp_port,
                                                 discover_rate=self.thymio.discover_rate,
                                                 refreshing_rate=self.thymio.refreshing_rate,
                                                 loop=self.loop)
            else:
                self.connection = Connection.serial(port=self.thymio.serial_port,
                                                    discover_rate=self.thymio.discover_rate,
                                                    refreshing_rate=self.thymio.refreshing_rate,
                                                    loop=self.loop)
            self.connection.on_connection_changed = on_connection_changed
            self.connection.on_variables_received = on_variables_received
            self.connection.on_user_event = on_user_event

            self.loop.run_forever()

    def __init__(self,
                 use_tcp=False,
                 serial_port=None,
                 host=None,
                 tcp_port=None,
                 on_connect=None,
                 on_disconnect=None,
                 refreshing_rate=0.1,
                 discover_rate=2,
                 loop=None):
        self.use_tcp = use_tcp
        self.serial_port = serial_port
        self.host = host
        self.tcp_port = tcp_port
        self.on_connect_cb = on_connect
        self.on_disconnect_cb = on_disconnect
        self.refreshing_rate = refreshing_rate
        self.discover_rate = discover_rate
        self.loop = loop or asyncio.get_event_loop()
        self.thymio_proxy = None
        self.variable_observers = {}
        self.user_event_listeners = {}

    def connect(self):
        """Connect to Thymio or dongle.
        """
        def thymio_thread():
            asyncio.set_event_loop(asyncio.new_event_loop())
            self.thymio_proxy = self._ThymioProxy(self)
            self.thymio_proxy.run()
        self.thread = threading.Thread(target=thymio_thread)
        self.thread.start()
        while self.thymio_proxy is None or len(self.thymio_proxy.nodes) == 0:
            time.sleep(0.1)

    def nodes(self):
        """Get set of ids of node currentlty connected.
        """
        return self.thymio_proxy.nodes if self.thymio_proxy else set()

    def first_node(self):
        """Get id of first node connected.
        """
        return next(iter(self.nodes()))

    def variables(self, node_id):
        """Get list of variable names.
        """
        node = self.thymio_proxy.connection.remote_nodes[node_id]
        return node.named_variables

    def variable_size(self, node_id, var_name):
        """Get the size of a variable.
        """
        node = self.thymio_proxy.connection.remote_nodes[node_id]
        return node.var_size[var_name]

    def variable_offset(self, node_id, var_name):
        """Get the offset (address) of a variable.
        """
        node = self.thymio_proxy.connection.remote_nodes[node_id]
        return node.var_offset[var_name]

    def __getitem__(self, key):
        class Node:
            def __init__(self_node, node_id):
                self_node.node_id = node_id
            def __getitem__(self_node, name):
                try:
                    val = self.thymio_proxy.connection.get_var_array(self_node.node_id, name)
                    return val if len(val) != 1 else val[0]
                except KeyError:
                    raise KeyError(name)
            def __setitem__(self_node, name, val):
                try:
                    if isinstance(val, list):
                        self.thymio_proxy.connection.set_var_array(self_node.node_id, name, val)
                    else:
                        self.thymio_proxy.connection.set_var(self_node.node_id, name, val)
                except KeyError:
                    raise KeyError(name)

        return Node(key)

    def set_variable_observer(self, node_id, observer):
        self.variable_observers[node_id] = observer

    def set_user_event_listener(self, node_id, listener):
        self.user_event_listeners[node_id] = listener

    def events(self, node_id):
        """Get list of event names.
        """
        node = self.thymio_proxy.connection.remote_nodes[node_id]
        return node.local_events

    def native_functions(self, node_id):
        """Get list of native function names and list of corresponding arg sizes.
        """
        node = self.thymio_proxy.connection.remote_nodes[node_id]
        return (
            node.native_functions,
            [node.native_functions_arg_sizes[f] for f in node.native_functions]
        )

    def run_asm(self, node_id: int, asm: str) -> None:
        """Assemble assembly code to bytecode, load it and run it.
        """
        # assemble program
        remote_node = self.thymio_proxy.connection.remote_nodes[node_id]
        a = Assembler(remote_node, asm)
        bc = a.assemble()
        # run it
        self.thymio_proxy.connection.set_bytecode(node_id, bc)
        self.thymio_proxy.connection.run(node_id)



# This file is part of thymiodirect.
# Copyright 2020 ECOLE POLYTECHNIQUE FEDERALE DE LAUSANNE,
# Miniature Mobile Robots group, Switzerland
# Author: Yves Piguet
#
# SPDX-License-Identifier: BSD-3-Clause

"""
Assembler for Aseba VM bytecode
Author: Yves Piguet, EPFL
"""

import thymiodirect
import re
from typing import Callable, Dict, List, Union


class Assembler:

    def __init__(self, remote_node: thymiodirect.connection.RemoteNode, src: str):
        """
        Construct a new Assembler object.

        Args:
            remote_node: Remote node used to predefine constants.
            src: Assembly source code.
        """

        self.remote_node = remote_node
        self.src = src

        self.instr = {
            "dc": {
                "num_args": -1
            },
            "equ": {
                "num_args": 1
            },
            "stop": {
                "code": [0x0000]
            },
            "push.s": {
                "num_args": 1
            },
            "push": {
                "num_args": 1
            },
            "load": {
                "num_args": 1
            },
            "store": {
                "num_args": 1
            },
            "load.ind": {
                "num_args": 2
            },
            "store.ind": {
                "num_args": 2
            },
            "neg": {
                "code": [0x7000]
            },
            "abs": {
                "code": [0x7001]
            },
            "bitnot": {
                "code": [0x7002]
            },
            "not": {
                # empty
            },
            "sl": {
                "code": [0x8000]
            },
            "asr": {
                "code": [0x8001]
            },
            "add": {
                "code": [0x8002]
            },
            "sub": {
                "code": [0x8003]
            },
            "mult": {
                "code": [0x8004]
            },
            "div": {
                "code": [0x8005]
            },
            "mod": {
                "code": [0x8006]
            },
            "bitor": {
                "code": [0x8007]
            },
            "bitxor": {
                "code": [0x8008]
            },
            "bitand": {
                "code": [0x8009]
            },
            "eq": {
                "code": [0x800a]
            },
            "ne": {
                "code": [0x800b]
            },
            "gt": {
                "code": [0x800c]
            },
            "ge": {
                "code": [0x800d]
            },
            "lt": {
                "code": [0x800e]
            },
            "le": {
                "code": [0x800f]
            },
            "or": {
                "code": [0x8010]
            },
            "and": {
                "code": [0x8011]
            },
            "jump": {
                "num_args": 1
            },
            "jump.if.not": {
                "num_args": 2
            },
            "do.jump.when.not": {
                "num_args": 2
            },
            "dont.jump.when.not": {
                "num_args": 2
            },
            "emit": {
                "num_args": 3
            },
            "callnat": {
                "num_args": 1
            },
            "callsub": {
                "num_args": 1
            },
            "ret": {
                "code": [0xe000]
            },
        }

        def resolve_symbol(a, defs: Dict[str, int], required: bool) -> int:

            def resolve_def(name: str) -> int:
                if not required:
                    return 0
                if re.match("^(0x[0-9a-f]+|[0-9]+)$", name, flags=re.I):
                    return int(name, 0)
                if name not in defs:
                    raise Exception(f'Unknown symbol "{name}"')
                return defs[name]

            if type(a) is str:
                # eval
                val = 0
                minus = False
                offset = 0
                while offset < len(a):
                    r = re.match("(\+|-|[._a-z0-9]+)", a[offset:], re.I)
                    if r is None:
                        raise Exception("Syntax error")
                    s = r.group()
                    if s == "+":
                        minus = False
                    elif s == "-":
                        minus = True
                    else:
                        val += -resolve_def(s) if minus else resolve_def(s)
                    offset += len(s)
                return val

            return a

        def def_to_code(instr: str) -> Callable:
            def register(fun):
                self.instr[instr]["to_code"] = fun
                return fun
            return register

        @def_to_code("dc")
        def to_code_dc(pc: int, args: List[Union[int, str]], label: str, defs: Dict[str, int], phase: int, line: int) -> List[int]:
            return [
                resolve_symbol(a, defs, phase == 1) & 0xffff
                for a in args
            ]

        @def_to_code("equ")
        def to_code_equ(pc, args, label, defs, phase, line):
            if label is None:
                raise Exception(f'No label for pseudo-instruction "equ" (line {line})')
            if defs is not None:
                defs[label] = resolve_symbol(args[0], defs, phase == 1)
                label = None
            return []

        @def_to_code("push.s")
        def to_code_push_s(pc, args, label, defs, phase, line):
            arg = resolve_symbol(args[0], defs, phase == 1)
            if arg >= 0x1000 or -arg > 0x1000:
                raise Exception(f"Small integer overflow (line {line})")
            return [0x1000 | arg & 0xfff]

        @def_to_code("push")
        def to_code_push(pc, args, label, defs, phase, line):
            arg = resolve_symbol(args[0], defs, phase == 1)
            return [0x2000, arg & 0xffff]

        @def_to_code("load")
        def to_code_load(pc, args, label, defs, phase, line):
            arg = resolve_symbol(args[0], defs, phase == 1)
            if arg < 0 or arg >= 0x1000:
                raise Exception(f"Data address out of range (line {line})")
            return [0x3000 | arg & 0xfff]

        @def_to_code("store")
        def to_code_store(pc, args, label, defs, phase, line):
            arg = resolve_symbol(args[0], defs, phase == 1)
            if arg < 0 or arg >= 0x1000:
                raise Exception(f"Data address out of range (line {line})")
            return [0x4000 | arg & 0xfff]

        @def_to_code("load.ind")
        def to_code_load_ind(pc, args, label, defs, phase, line):
            arg = resolve_symbol(args[0], defs, phase == 1)
            if arg < 0 or arg >= 0x1000:
                raise Exception(f"Data address out of range (line {line})")
            size_arg = resolve_symbol(args[1], defs, phase == 1)
            return [0x5000 | arg & 0xfff, size_arg & 0xffff]

        @def_to_code("store.ind")
        def to_code_store_ind(pc, args, label, defs, phase, line):
            arg = resolve_symbol(args[0], defs, phase == 1)
            if arg < 0 or arg >= 0x1000:
                raise Exception(f"Data address out of range (line {line})")
            size_arg = resolve_symbol(args[1], defs, phase == 1)
            return [0x6000 | arg & 0xfff, size_arg & 0xffff]

        @def_to_code("not")
        def to_code_not(pc, args, label, defs, phase, line):
            raise Exception(f'Unary "not" not implemented in the VM (line {line})')

        @def_to_code("jump")
        def to_code_jump(pc, args, label, defs, phase, line):
            arg = resolve_symbol(args[0], defs, phase == 1)
            return [0x9000 | (arg - pc) & 0xfff]

        @def_to_code("jump.if.not")
        def to_code_jump_if_not(pc, args, label, defs, phase, line):
            test_instr = self.instr[args[0]] if args[0] in self.instr else None
            if (test_instr is None
                or "code" not in test_instr
                or len(test_instr["code"]) != 1
                or (test_instr["code"][0] & 0xf000) != 0x8000):
                raise Exception(f'Unknown op "{args[0]}" for jump.if.not (line {line})')
            arg = resolve_symbol(args[1], defs, phase == 1)
            return [0xa000 | (test_instr["code"][0] & 0xff), (arg - pc) & 0xffff]

        @def_to_code("do.jump.when.not")
        def to_code_do_jump_when_not(pc, args, label, defs, phase, line):
            test_instr = self.instr[args[0]] if args[0] in self.instr else None
            if (test_instr is None
                or "code" not in test_instr
                or len(test_instr["code"]) != 1
                or (test_instr["code"][0] & 0xf000) != 0x8000):
                raise Exception(f'Unknown op "{args[0]}" for do.jump.when.not (line {line})')
            arg = resolve_symbol(args[1], defs, phase == 1)
            return [0xa100 | (test_instr["code"][0] & 0xff), (arg - pc) & 0xffff]

        @def_to_code("dont.jump.when.not")
        def to_code_do_jump_when_not(pc, args, label, defs, phase, line):
            test_instr = self.instr[args[0]] if args[0] in self.instr else None
            if (test_instr is None
                or "code" not in test_instr
                or len(test_instr["code"]) != 1
                or (test_instr["code"][0] & 0xf000) != 0x8000):
                raise Exception(f'Unknown op "{args[0]}" for dont.jump.when.not (line {line})')
            arg = resolve_symbol(args[1], defs, phase == 1)
            return [0xa300 | (test_instr["code"][0] & 0xff), (arg - pc) & 0xffff]

        @def_to_code("emit")
        def to_code_emit(pc, args, label, defs, phase, line):
            id = resolve_symbol(args[0], defs, phase == 1)
            if id < 0 or id >= 0x1000:
                raise Exception(f"Event id out of range (line {line})")
            addr = resolve_symbol(args[1], defs, phase == 1)
            size = resolve_symbol(args[2], defs, phase == 1)
            return [0xb000 | id & 0xfff, addr & 0xffff, size & 0xffff]

        @def_to_code("callnat")
        def to_code_callnat(pc, args, label, defs, phase, line):
            arg = resolve_symbol(args[0], defs, phase == 1)
            if arg < 0 or arg >= 0x1000:
                raise Exception(f"Native call id out of range (line {line})")
            return [0xc000 | arg & 0xfff]

        @def_to_code("callsub")
        def to_code_callsub(pc, args, label, defs, phase, line):
            arg = resolve_symbol(args[0], defs, phase == 1)
            if arg < 0 or arg >= 0x1000:
                raise Exception(f"Subroutine address out of range (line {line})")
            return [0xd000 | arg & 0xfff]

    def node_definitions(self) -> None:
        """Create definition dict based on node variables and native functions.
        """

        defs = {}

        # variables
        for name in self.remote_node.named_variables:
            defs[name] = self.remote_node.var_offset[name]
        defs["_userdata"] = self.remote_node.var_total_size
        defs["_topdata"] = self.remote_node.max_var_size

        # local events
        defs["_ev.init"] = 0xffff
        for i in range(len(self.remote_node.local_events)):
            defs["_ev." + self.remote_node.local_events[i]] = 0xfffe - i

        # native functions
        for i in range(len(self.remote_node.native_functions)):
            defs["_nf." + self.remote_node.native_functions[i]] = i

        return defs

    def assemble(self) -> List[int]:
        """Assemble to bytecode.
        """

        lines = self.src.split("\n")
        defs = self.node_definitions()

        re_blank = re.compile(r"^\s*(;.*)?$")
        re_label = re.compile(r"^\s*([\w_.]+:)\s*(;.*)?$")
        re_instr = re.compile(r"^\s*([\w_.]+:)?\s*([a-z0-9.]+)([-a-zA-Z0-9\s._,+=]*)(;.*)?$")
        re_number = re.compile(r"^(-?[0-9]+|0x[0-9a-fA-F]+)$")

        for phase in (0, 1):
            bytecode = []
            label = None
            for i, line in enumerate(lines):
                if re_blank.match(line):
                    # blank or comment (ignore)
                    continue

                r = re_label.match(line)
                if r:
                    # label without instr
                    label = r[1][0:-1]
                    defs[label] = len(bytecode)
                    continue

                r = re_instr.match(line)
                if r:
                    if r[1]:
                        label = r[1][0:-1]
                        defs[label] = len(bytecode)
                    instr_name = r[2]
                    instr_args = r[3].strip()
                    if instr_name:
                        if len(instr_args) > 0:
                            args_split = [
                                a.strip()
                                for a1 in instr_args.split(",")
                                for a in a1.strip().split()
                            ]
                        else:
                            args_split = []

                    if instr_name not in self.instr:
                        raise Exception(f"Unknown instruction {instr_name} (line {i+1})")
                    instr = self.instr[instr_name]
                    if "code" in instr:
                        bytecode += instr["code"]
                    elif "to_code" in instr:
                        args = [
                            int(a, 0) if re_number.match(a) else a
                            for a in args_split
                        ]
                        bytecode += instr["to_code"](len(bytecode), args, label, defs, phase, i + 1)
                    if label is not None and defs[label] != len(bytecode):
                        label = None
                    continue

                print("line", line)
                raise Exception(f"Syntax error (line {i+1})")

        return bytecode

def test(remote_node=None):
    if remote_node is None:
        remote_node = thymiodirect.connection.RemoteNode()

    src = """foo:
    equ 5

    dc end_toc
    dc _ev.init, init
end_toc:

init:
    push foo+1
    stop
"""
    a = Assembler(remote_node, src)
    bc = a.assemble()
    print(src)
    for i, c in enumerate(bc):
        print(f"{i:4d} {c:0=4x}")



# This file is part of thymiodirect.
# Copyright 2020 ECOLE POLYTECHNIQUE FEDERALE DE LAUSANNE,
# Miniature Mobile Robots group, Switzerland
# Author: Yves Piguet
#
# SPDX-License-Identifier: BSD-3-Clause

# Messages defined for Aseba communication protocol

import uuid

class Message:
    """Aseba message data.
    """

    # v5
    ID_FIRST_ASEBA_ID = 0x8000
    ID_BOOTLOADER_RESET = 0x8000
    ID_BOOTLOADER_READ_PAGE = 0x8001
    ID_BOOTLOADER_WRITE_PAGE = 0x8002
    ID_BOOTLOADER_PAGE_DATA_WRITE = 0x8003
    ID_BOOTLOADER_DESCRIPTION = 0x8004
    ID_BOOTLOADER_PAGE_DATA_READ = 0x8005
    ID_BOOTLOADER_ACK = 0x8006
    ID_DESCRIPTION = 0x9000
    ID_NAMED_VARIABLE_DESCRIPTION = 0x9001
    ID_LOCAL_EVENT_DESCRIPTION = 0x9002
    ID_NATIVE_FUNCTION_DESCRIPTION = 0x9003
    ID_VARIABLES = 0x9005
    ID_EXECUTION_STATE_CHANGED = 0x900a
    ID_NODE_PRESENT = 0x900c
    ID_DEVICE_INFO = 0x900d
    ID_CHANGED_VARIABLES = 0x900e
    ID_GET_DESCRIPTION = 0xa000
    ID_SET_BYTECODE = 0xa001
    ID_RESET = 0xa002
    ID_RUN = 0xa003
    ID_PAUSE = 0xa004
    ID_STEP = 0xa005
    ID_STOP = 0xa006
    ID_GET_EXECUTION_STATE = 0xa007
    ID_BREAKPOINT_SET = 0xa008
    ID_BREAKPOINT_CLEAR = 0xa009
    ID_BREAKPOINT_CLEAR_ALL = 0xa00a
    ID_GET_VARIABLES = 0xa00b
    ID_SET_VARIABLES = 0xa00c
    ID_GET_NODE_DESCRIPTION = 0xa010
    ID_LIST_NODES = 0xa011
    # v6
    ID_GET_DEVICE_INFO = 0xa012
    ID_SET_DEVICE_INFO = 0xa013
    # v7
    ID_GET_CHANGED_VARIABLES = 0xa014
    # v8
    ID_GET_NODE_DESCRIPTION_FRAGMENT = 0xa015

    PROTOCOL_VERSION = 5

    DEVICE_INFO_UUID = 1
    DEVICE_INFO_NAME = 2
    DEVICE_INFO_THYMIO2_RF_SETTINGS = 3

    def __init__(self, id, source_node, payload):
        self.id = id
        self.source_node = source_node
        self.payload = payload

    def get_uint8(self, offset):
        """Get an unsigned 8-bit integer in the payload.
        """
        return self.payload[offset], offset + 1

    def get_uint16(self, offset):
        """Get an unsigned 16-bit integer in the payload.
        """
        return self.payload[offset] + 256 * self.payload[offset + 1], offset + 2

    def get_string(self, offset):
        """Get a string in the payload.
        """
        len = self.payload[offset]
        str = self.payload[offset + 1:offset + 1 + len]
        return str.decode('utf-8'), offset + 1 + len

    @staticmethod
    def uint16_to_bytes(word):
        """Convert an unsigned 16-bit integer to bytes.
        """
        return bytes([word & 0xff, (word & 0xff00) // 256])

    @staticmethod
    def uint16array_to_bytes(a):
        """Convert an array of unsigned 16-bit integer to bytes.
        """
        bytes = b""
        for word in a:
            bytes += Message.uint16_to_bytes(word)
        return bytes

    def decode(self):
        """Decode message properties from its payload.
        """
        if self.id == Message.ID_DESCRIPTION:
            self.node_name, offset = self.get_string(0)
            self.protocol_version, offset = self.get_uint16(offset)
            self.bytecode_size, offset = self.get_uint16(offset)
            self.stack_size, offset = self.get_uint16(offset)
            self.max_var_size, offset = self.get_uint16(offset)
            self.num_named_var, offset = self.get_uint16(offset)
            self.num_local_events, offset = self.get_uint16(offset)
            self.num_native_fun, offset = self.get_uint16(offset)
        elif self.id == Message.ID_NAMED_VARIABLE_DESCRIPTION:
            self.var_size, offset = self.get_uint16(0)
            self.var_name, offset = self.get_string(offset)
        elif self.id == Message.ID_LOCAL_EVENT_DESCRIPTION:
            self.event_name, offset = self.get_string(0)
            self.description, offset = self.get_string(offset)
        elif self.id == Message.ID_NATIVE_FUNCTION_DESCRIPTION:
            self.fun_name, offset = self.get_string(0)
            self.description, offset = self.get_string(offset)
            num_params, offset = self.get_uint16(offset)
            self.param_names = []
            self.param_sizes = []
            for i in range(num_params):
                size, offset = self.get_uint16(offset)
                name, offset = self.get_string(offset)
                self.param_names.append(name)
                self.param_sizes.append(size)
        elif self.id == Message.ID_VARIABLES:
            self.var_offset, offset = self.get_uint16(0)
            self.var_data = []
            for i in range(len(self.payload) // 2 - 1):
                word, offset = self.get_uint16(offset)
                self.var_data.append(word)
        elif self.id == Message.ID_EXECUTION_STATE_CHANGED:
            self.pc, offset = self.get_uint16(0)
            self.flags, offset = self.get_uint16(offset)
            self.event_active = (self.flags & 1) != 0
            self.step_by_step = (self.flags & 2) != 0
            self.event_running = (self.flags & 4) != 0
        elif self.id == Message.ID_NODE_PRESENT:
            self.version, offset = self.get_uint16(0)
        elif self.id == Message.ID_DEVICE_INFO:
            self.device_info, offset = self.get_uint8(0)
            if self.device_info == Message.DEVICE_INFO_NAME:
                self.device_name, offset = self.get_string(offset)
            elif self.device_info == Message.DEVICE_INFO_UUID:
                data_len, offset = self.get_uint8(offset)
                data = self.payload[offset:offset + data_len]
                self.device_uuid = str(uuid.UUID(bytes=data))
            elif self.device_info == Message.DEVICE_INFO_THYMIO2_RF_SETTINGS:
                data_len, offset = self.get_uint8(offset)
                if data_len == 6:
                    self.network_id, offset = self.get_uint16(offset)
                    self.node_id, offset = self.get_uint16(offset)
                    self.channel, offset = self.get_uint16(offset)
        elif self.id == Message.ID_SET_BYTECODE:
            self.target_node_id, offset = self.get_uint16(0)
            self.bc_offset, offset = self.get_uint16(offset)
            val = []
            for i in range(4, len(self.payload), 2):
                instr, offset = self.get_uint16(offset)
                val.append(instr)
            self.bc = val
        elif (self.id == Message.ID_BREAKPOINT_CLEAR_ALL
              or self.id == Message.ID_RESET
              or self.id == Message.ID_RUN
              or self.id == Message.ID_PAUSE
              or self.id == Message.ID_STEP
              or self.id == Message.ID_STOP
              or self.id == Message.ID_GET_EXECUTION_STATE):
            self.target_node_id, offset = self.get_uint16(0)
        elif (self.id == Message.ID_BREAKPOINT_SET
              or self.id == Message.ID_BREAKPOINT_CLEAR):
            self.target_node_id, offset = self.get_uint16(0)
            self.pc, offset = self.get_uint16(offset)
        elif self.id == Message.ID_GET_VARIABLES:
            self.target_node_id, offset = self.get_uint16(0)
            self.var_offset, offset = self.get_uint16(offset)
            self.var_count, offset = self.get_uint16(offset)
        elif self.id == Message.ID_SET_VARIABLES:
            self.target_node_id, offset = self.get_uint16(0)
            self.var_offset, offset = self.get_uint16(offset)
            val = []
            for i in range(4, len(self.payload), 2):
                v, offset = self.get_uint16(offset)
                val.append(v)
            self.var_val = val
        elif self.id == Message.ID_LIST_NODES:
            self.version, offset = self.get_uint16(0)
        elif self.id == Message.ID_GET_NODE_DESCRIPTION_FRAGMENT:
            self.version, offset = self.get_uint16(0)
            self.fragment, offset = self.get_uint16(offset)
        elif self.id < Message.ID_FIRST_ASEBA_ID:
            val = []
            offset = 0
            for i in range(0, len(self.payload), 2):
                v, offset = self.get_uint16(offset)
                val.append(v)
            self.user_event_arg = val

    def serialize(self):
        """Serialize message to bytes.
        """
        return (self.uint16_to_bytes(len(self.payload))
                + self.uint16_to_bytes(self.source_node)
                + self.uint16_to_bytes(self.id)
                + self.payload)

    @staticmethod
    def id_to_str(id):
        """Convert message id to its name string.
        """
        try:
            return {
                Message.ID_DESCRIPTION: "DESCRIPTION",
                Message.ID_NAMED_VARIABLE_DESCRIPTION: "ID_NAMED_VARIABLE_DESCRIPTION",
                Message.ID_LOCAL_EVENT_DESCRIPTION: "ID_LOCAL_EVENT_DESCRIPTION",
                Message.ID_NATIVE_FUNCTION_DESCRIPTION: "ID_NATIVE_FUNCTION_DESCRIPTION",
                Message.ID_VARIABLES: "ID_VARIABLES",
                Message.ID_EXECUTION_STATE_CHANGED: "ID_EXECUTION_STATE_CHANGED",
                Message.ID_NODE_PRESENT: "ID_NODE_PRESENT",
                Message.ID_GET_DESCRIPTION: "ID_GET_DESCRIPTION",
                Message.ID_SET_BYTECODE: "ID_SET_BYTECODE",
                Message.ID_RESET: "ID_RESET",
                Message.ID_RUN: "ID_RUN",
                Message.ID_PAUSE: "ID_PAUSE",
                Message.ID_STEP: "ID_STEP",
                Message.ID_STOP: "ID_STOP",
                Message.ID_GET_EXECUTION_STATE: "ID_GET_EXECUTION_STATE",
                Message.ID_BREAKPOINT_SET: "ID_BREAKPOINT_SET",
                Message.ID_BREAKPOINT_CLEAR: "ID_BREAKPOINT_CLEAR",
                Message.ID_BREAKPOINT_CLEAR_ALL: "ID_BREAKPOINT_CLEAR_ALL",
                Message.ID_GET_VARIABLES: "ID_GET_VARIABLES",
                Message.ID_SET_VARIABLES: "ID_SET_VARIABLES",
                Message.ID_GET_NODE_DESCRIPTION: "ID_GET_NODE_DESCRIPTION",
                Message.ID_LIST_NODES: "ID_LIST_NODES",
                Message.ID_GET_DEVICE_INFO: "ID_GET_DEVICE_INFO",
                Message.ID_SET_DEVICE_INFO: "ID_SET_DEVICE_INFO",
                Message.ID_GET_CHANGED_VARIABLES: "ID_GET_CHANGED_VARIABLES",
                Message.ID_GET_NODE_DESCRIPTION_FRAGMENT: "ID_GET_NODE_DESCRIPTION_FRAGMENT",
            }[id]
        except KeyError as error:
            return f"ID {id}"

    def __str__(self):
        str = f"Message id={self.id_to_str(self.id)} src={self.source_node}"
        if self.id == Message.ID_DESCRIPTION:
            str += f" name={self.node_name}"
            str += f" vers={self.protocol_version}"
            str += f" bc_size={self.bytecode_size}"
            str += f" stack_size={self.stack_size}"
            str += f" max_var_size={self.max_var_size}"
            str += f" #var={self.num_named_var}"
            str += f" #ev={self.num_local_events}"
            str += f" #nat={self.num_native_fun}"
        elif self.id == Message.ID_NAMED_VARIABLE_DESCRIPTION:
            str += f" name={self.var_name} size={self.var_size}"
        elif self.id == Message.ID_LOCAL_EVENT_DESCRIPTION:
            str += f" name={self.event_name} descr={self.description}"
        elif self.id == Message.ID_NATIVE_FUNCTION_DESCRIPTION:
            str += f" name={self.fun_name} descr={self.description} p=("
            for i in range(len(self.param_names)):
                str += f"{self.param_names[i]}[{self.param_sizes[i] if self.param_sizes[i] != 65535 else '?'}],"
            str += ")"
        elif self.id == Message.ID_VARIABLES:
            str += f" offset={self.var_offset} data=("
            for word in self.var_data:
                str += f"{word},"
            str += ")"
        elif self.id == Message.ID_EXECUTION_STATE_CHANGED:
            str += f" pc={self.pc} event_active={self.event_active} step_by_step={self.step_by_step} event_running={self.event_running}"

        elif self.id == Message.ID_NODE_PRESENT:
            str += f" version={self.version}"
        return str



# This file is part of thymiodirect.
# Copyright 2020 ECOLE POLYTECHNIQUE FEDERALE DE LAUSANNE,
# Miniature Mobile Robots group, Switzerland
# Author: Yves Piguet
#
# SPDX-License-Identifier: BSD-3-Clause

from setuptools import setup

with open("help.md", "r") as fh:
    long_description = fh.read()

setup(
    name='thymiodirect',
    version='0.1.1',
    author='Yves Piguet',
    packages=['thymiodirect'],
    description='Communication with Thymio II robot via serial port or TCP',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/epfl-mobots/thymio-python',
    install_requires=[
        'pyserial'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Intended Audience :: Education",
        "Framework :: AsyncIO"
    ],
    python_requires='>=3.6',
)



# This file is part of thymiodirect.
# Copyright 2020 ECOLE POLYTECHNIQUE FEDERALE DE LAUSANNE,
# Miniature Mobile Robots group, Switzerland
# Author: Yves Piguet
#
# SPDX-License-Identifier: BSD-3-Clause

"""
Communication with Thymio via serial port or tcp
Author: Yves Piguet, EPFL
"""

import asyncio
import threading
import time
from typing import List, Optional

from thymiodirect.message import Message


class InputThread(threading.Thread):
    """Thread which reads messages asynchronously.
    """

    def __init__(self, io, loop=None, handle_msg=None):
        threading.Thread.__init__(self)
        self.running = True
        self.io = io
        self.loop = loop
        self.handle_msg = handle_msg

    def terminate(self) -> None:
        self.running = False

    def read_uint16(self) -> int:
        """Read an unsigned 16-bit number.
        """
        b = self.io.read(2)
        if len(b) == 0:
            raise TimeoutError()
        elif len(b) == 1:
            b1 = self.io.read(1)
            if len(b1) == 0:
                raise TimeoutError()
            return b[0] + 256 * b1[0]
        else:
            return b[0] + 256 * b[1]

    def read_message(self) -> Message:
        """Read a complete message.
        """
        payload_len = self.read_uint16()
        source_node = self.read_uint16()
        id = self.read_uint16()
        payload = self.io.read(payload_len)
        msg = Message(id, source_node, payload)
        return msg

    def run(self) -> None:
        """Input thread code.
        """
        while self.running:
            try:
                msg = self.read_message()
                msg.decode()
                if self.loop and self.handle_msg:
                    asyncio.ensure_future(self.handle_msg(msg), loop=self.loop)
            except TimeoutError:
                pass


class RemoteNode:
    """Remote node description and state.
    """

    def __init__(self, node_id: Optional[int] = None, version: Optional[int] = None):
        self.node_id = node_id
        self.version = version
        self.device_name = None
        self.device_uuid = None
        self.rf_network_id = None
        self.rf_node_id = None
        self.rf_channel = None
        self.last_msg_time = 0  # time.time()
        self.handshake_done = False
        self.name = None
        self.max_var_size = None
        self.num_named_var = None
        self.num_local_events = None
        self.num_native_fun = None
        self.var_total_size = 0
        self.named_variables = []  # names
        self.var_offset = {}  # indexed by name
        self.var_size = {}  # indexed by name
        self.var_data = []
        self.expected_var_end = 0  # beyond last var requested by ID_GET_VARIABLES
        self.var_received = False  # True if last set_var_data reached expected_var_end
        self.local_events = []  # names
        self.native_functions = []  # names
        self.native_functions_arg_sizes = {}  # indexed by name

    def __str__(self) -> str:
        return f"RemoteNode node_id={self.node_id} uuid={self.device_uuid}"

    def __repr__(self) -> str:
        return str(self)

    def add_var(self, name: str, size: int) -> None:
        """Add the definition of a variable.
        """
        self.named_variables.append(name)
        self.var_offset[name] = self.var_total_size
        self.var_size[name] = size
        self.var_total_size += size

    def reset_var_data(self) -> None:
        """Reset the variable data to 0.
        """
        self.var_data = [0 for i in range(self.var_total_size)]

    def get_var(self, name: str, index: int = 0) -> int:
        """Get the value of a scalar variable or an item in an array variable.
        """
        return self.var_data[self.var_offset[name] + index]

    def get_var_array(self, name: str) -> List[int]:
        """Get the value of an array variable.
        """
        if name not in self.var_offset:
            raise KeyError(name)
        offset = self.var_offset[name]
        return self.var_data[offset:offset + self.var_size[name]]

    def set_var(self, name: str, val: int, index: Optional[int] = 0) -> None:
        """Set the value of a scalar variable or an item in an array variable.
        """
        self.var_data[self.var_offset[name] + index] = val

    def set_var_array(self, name: str, val: List[int]) -> None:
        """Set the value of an array variable.
        """
        offset = self.var_offset[name]
        self.var_data[offset:offset + len(val)] = val

    def set_var_data(self, offset: int, data: List[int]) -> int:
        """Set values in the variable data array.
        """
        self.var_data[offset:offset + len(data)] = data
        self.var_received = offset + len(data) >= self.expected_var_end


class Connection:
    """Connection to one or multiple devices.
    """

    class ThymioConnectionError(OSError):
        pass

    def __init__(self,
                 io,
                 host_node_id=1, refreshing_rate=None, discover_rate=None,
                 debug=False,
                 loop=None):
        self.loop = loop or asyncio.new_event_loop()
        self.terminating = False
        self.io = io
        self.debug = debug
        self.timeout = 3
        self.host_node_id = host_node_id
        self.auto_handshake = False
        self.remote_node_set = set()  # set of id of nodes with handshake done
        self.remote_nodes = {}  # key: node_id

        self.input_lock = threading.Lock()
        self.input_thread = InputThread(self.io, self.loop,
                                        lambda msg: self.handle_message(msg))
        self.input_thread.start()

        self.output_lock = threading.Lock()
        self.refreshing_timeout = None
        self.refreshing_triggers = []   # threading.Event
        if refreshing_rate is not None:
            self.set_refreshing_rate(refreshing_rate)

        # callback for (dis)connection
        # async fun(node_id, connect)
        self.on_connection_changed = None

        # callback for notification that variables have been received
        # async fun(node_id)
        self.on_variables_received = None

        # callback for notification about execution state
        # async fun(node_id, pc, flags)
        self.on_execution_state_changed = None

        # callback for notification that an event has been emitted
        # async fun(node_id, event_id, event_args)
        self.on_user_event = None

        # discover coroutine
        if discover_rate is not None:
            async def discover():
                while not self.terminating:
                    self.handshake()
                    await asyncio.sleep(discover_rate)
            self.loop.create_task(discover())

    def close(self) -> None:
        """Close connection.
        """
        if not self.io.closed:
            self.io.close()

    def shutdown(self) -> None:
        """Shutdown everything.
        """
        self.terminating = True
        self.input_thread.terminate()

    def run_forever(self) -> None:
        """Run asyncio loop forever.
        """
        self.loop.run_forever()

    def __del__(self) -> None:
        self.shutdown()
        self.close()

    def __enter__(self) -> None:
        return self

    def __exit__(self, type, value, traceback) -> None:
        self.close()

    @staticmethod
    def serial_ports() -> List[str]:
        """Get the list of serial ports for the current platform.
        """
        import sys
        import os
        if sys.platform == "linux":
            devices = [
                "/dev/" + filename
                for filename in os.listdir("/dev")
                if filename.startswith("ttyACM")
            ]
        elif sys.platform == "darwin":
            devices = [
                "/dev/" + filename
                for filename in os.listdir("/dev")
                if filename.startswith("cu.usb")
            ]
        elif sys.platform == "win32":
            import subprocess, re
            mode_output = subprocess.check_output("mode", shell=True).decode()
            devices = [
                re.search(r"(COM\d+):", line).groups()[0]
                for line in mode_output.split("\n")
                if re.search(r"(COM\d+):", line)
            ]
        else:
            raise Connection.ThymioConnectionError("Unsupported platform")
        return devices

    @staticmethod
    def serial_default_port() -> str:
        """Get the name of the default Thymio serial port for the current
        platform.
        """
        import sys
        devices = Connection.serial_ports()
        if len(devices) < 1:
            raise Connection.ThymioConnectionError("No serial device for Thymio found")
        if sys.platform == "win32":
            return devices[len(devices) - 1]
        else:
            return devices[0]

    @staticmethod
    def serial(port: Optional[str] = None, **kwargs) -> "Connection":
        """Create Thymio object with a serial connection.
        """
        import serial  # pip3 install pyserial
        if port is None:
            port = Connection.serial_default_port()
        th = Connection(serial.Serial(port, timeout=1), **kwargs)
        return th

    @staticmethod
    def tcp(host: Optional[str] = "127.0.0.1", port: Optional[int] = 33333, **kwargs) -> "Connection":
        """Create Thymio object with a TCP connection.
        """
        import socket, io

        class TCPClientIO(io.RawIOBase):

            def __init__(self, host, port):
                self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                self.socket.connect((host, port))

            def read(self, n):
                return self.socket.recv(n)

            def write(self, b):
                self.socket.sendall(b)

        s = TCPClientIO(host, port)
        th = Connection(s, **kwargs)
        return th

    @staticmethod
    def null(host_node_id: Optional[int] = 1, **kwargs) -> "Connection":
        """Create Thymio object without connection.
        """
        import io

        class NullIO(io.RawIOBase):

            def read(self, n):
                return None

            def write(self, b):
                pass

        return Connection(NullIO(), host_node_id)

    def handshake(self) -> None:
        self.auto_handshake = True
        self.list_nodes()

    def wait_for_handshake(self, n: Optional[int] = 1, timeout: Optional[int] = 5) -> None:
        """Wait until n remote nodes have finished handshake
        """
        if len(self.remote_node_set) < n and not self.auto_handshake:
            self.handshake()
        count = timeout // 0.1
        while len(self.remote_node_set) < n:
            time.sleep(0.1)
            count -= 1
            if count < 0:
                raise TimeoutError()

    def one_remote_node_id(self) -> int:
        """Get the node id of one of the connected nodes.
        """
        return next(iter(self.remote_node_set))

    def set_refreshing_rate(self, rate: float) -> None:
        """Change the auto-refresh rate to update variables.
        """
        self.refreshing_timeout = rate
        if rate is not None:
            # refresh now
            for event in self.refreshing_triggers:
                event.set()

    async def handle_message(self, msg: Message) -> None:
        """Handle an input message.
        """
        if self.debug:
            print("<", msg)
        source_node = msg.source_node
        if msg.id == Message.ID_NODE_PRESENT:
            will_do_handshake = False
            with self.input_lock:
                if source_node not in self.remote_nodes:
                    self.remote_nodes[source_node] = RemoteNode(source_node,
                                                                msg.version)
                    will_do_handshake = self.auto_handshake
            if will_do_handshake:
                if msg.version >= 6:
                    self.get_device_info(source_node)
                self.get_node_description(source_node)
        elif msg.id == Message.ID_DEVICE_INFO:
            change = False
            with self.input_lock:
                remote_node = self.remote_nodes[source_node]
                if msg.device_info == Message.DEVICE_INFO_NAME:
                    if remote_node.device_name != msg.device_name:
                        remote_node.device_name = msg.device_name
                        change = True
                elif msg.device_info == Message.DEVICE_INFO_UUID:
                    if remote_node.device_uuid != msg.device_uuid:
                        remote_node.device_uuid = msg.device_uuid
                        change = True
                elif msg.device_info == Message.DEVICE_INFO_THYMIO2_RF_SETTINGS:
                    if (remote_node.rf_network_id != msg.network_id
                        or remote_node.rf_node_id != msg.node_id
                        or remote_node.rf_channel != msg.channel):
                        remote_node.rf_network_id = msg.network_id
                        remote_node.rf_node_id = msg.node_id
                        remote_node.rf_channel = msg.channel
                        change = True
        elif msg.id == Message.ID_DESCRIPTION:
            with self.input_lock:
                self.remote_nodes[source_node].name = msg.node_name
                self.remote_nodes[source_node].bytecode_size = msg.bytecode_size
                self.remote_nodes[source_node].stack_size = msg.stack_size
                self.remote_nodes[source_node].max_var_size = msg.max_var_size
                self.remote_nodes[source_node].num_named_var = msg.num_named_var
                self.remote_nodes[source_node].num_local_events = msg.num_local_events
                self.remote_nodes[source_node].num_native_fun = msg.num_native_fun
        elif msg.id == Message.ID_NAMED_VARIABLE_DESCRIPTION:
            with self.input_lock:
                remote_node = self.remote_nodes[source_node]
                remote_node.add_var(msg.var_name, msg.var_size)
                if len(remote_node.named_variables) >= remote_node.num_named_var:
                    # all variables are known, can start refreshing
                    async def do_refresh():
                        while not self.terminating:
                            if self.refreshing_timeout is None:
                                await asyncio.sleep(0.1)
                            else:
                                await asyncio.sleep(self.refreshing_timeout)
                                self.get_variables(source_node)
                            # assume disconnection upon timeout
                            current_time = time.time()
                            terminating_nodes = set()
                            for node_id in self.remote_node_set:
                                with self.input_lock:
                                    remote_node = self.remote_nodes[node_id]
                                    if current_time - remote_node.last_msg_time > self.timeout:
                                        terminating_nodes.add(node_id)
                            for node_id in terminating_nodes:
                                self.remote_node_set.remove(node_id)
                                if self.on_connection_changed:
                                    await self.on_connection_changed(node_id, False)
                                del self.remote_nodes[node_id]
                        self.loop.stop()
                    self.loop.create_task(do_refresh())
        elif msg.id == Message.ID_VARIABLES:
            with self.input_lock:
                remote_node = self.remote_nodes[source_node]
                remote_node.set_var_data(msg.var_offset, msg.var_data)
            if self.on_variables_received and remote_node.var_received:
                await self.on_variables_received(source_node)
        elif msg.id == Message.ID_NATIVE_FUNCTION_DESCRIPTION:
            with self.input_lock:
                remote_node = self.remote_nodes[source_node]
                remote_node.native_functions.append(msg.fun_name)
                remote_node.native_functions_arg_sizes[msg.fun_name] = msg.param_sizes
            if len(remote_node.native_functions) >= remote_node.num_native_fun:
                # all messages sent as reply to GET_NODE_DESCRIPTION received
                self.remote_nodes[source_node].handshake_done = True
                if source_node not in self.remote_node_set:
                    self.remote_node_set.add(source_node)
                    if self.on_connection_changed:
                        await self.on_connection_changed(source_node, True)
        elif msg.id == Message.ID_LOCAL_EVENT_DESCRIPTION:
            with self.input_lock:
                remote_node = self.remote_nodes[source_node]
                remote_node.local_events.append(msg.event_name)
        elif msg.id == Message.ID_EXECUTION_STATE_CHANGED:
            if self.on_execution_state_changed:
                await self.on_execution_state_changed(source_node, msg.pc, msg.flags)
        elif msg.id < Message.ID_FIRST_ASEBA_ID:
            # user event sent by emit
            if self.on_user_event:
                await self.on_user_event(source_node, msg.id, msg.user_event_arg)
        self.remote_nodes[source_node].last_msg_time = time.time()

    def uuid_to_node_id(self, uuid: str) -> int:
        """Get node id from device uuid.
        """
        for node in self.remote_nodes.values():
            if node.device_uuid == uuid:
                return node.node_id

    def send(self, msg: Message) -> None:
        """Send a message.
        """
        with self.output_lock:
            if self.debug:
                print(">", msg)
            self.io.write(msg.serialize())

    def get_target_node_var_total_size(self, target_node_id):
        """Get the total size of variables.
        """
        with self.input_lock:
            return self.remote_nodes[target_node_id].var_total_size

    def list_nodes(self):
        """Send a LIST_NODES message.
        """
        payload = Message.uint16array_to_bytes([Message.PROTOCOL_VERSION])
        msg = Message(Message.ID_LIST_NODES, self.host_node_id, payload)
        self.send(msg)

    def get_node_description(self, target_node_id):
        """Send a GET_NODE_DESCRIPTION message.
        """
        payload = Message.uint16array_to_bytes([
            target_node_id,
            Message.PROTOCOL_VERSION
        ])
        msg = Message(Message.ID_GET_NODE_DESCRIPTION, self.host_node_id, payload)
        self.send(msg)

    def get_node_description_fragment(self, target_node_id, fragment):
        """Send a ID_GET_NODE_DESCRIPTION_FRAGMENT message.
        """
        payload = Message.uint16array_to_bytes([
            target_node_id,
            Message.PROTOCOL_VERSION,
            fragment
        ])
        msg = Message(Message.ID_GET_NODE_DESCRIPTION_FRAGMENT, self.host_node_id, payload)
        self.send(msg)

    def get_variables(self, target_node_id, chunk_offset=0, chunk_length=None):
        """Send a GET_VARIABLES message.
        """
        if target_node_id is not None:
            if chunk_length is None:
                chunk_length = (self.get_target_node_var_total_size(target_node_id)
                                - chunk_offset)
            payload = Message.uint16array_to_bytes([
                target_node_id,
                chunk_offset,
                chunk_length
            ])
            msg = Message(Message.ID_GET_VARIABLES, self.host_node_id, payload)
            self.remote_nodes[target_node_id].expected_var_end = chunk_offset + chunk_length
            self.send(msg)

    def set_variables(self, target_node_id, chunk_offset, chunk):
        """Send a SET_VARIABLES message.
        """
        payload = Message.uint16array_to_bytes([
            target_node_id,
            chunk_offset
        ] + chunk)
        msg = Message(Message.ID_SET_VARIABLES, self.host_node_id, payload)
        self.send(msg)

    def variable_description(self, target_node_id):
        """Get an array with the description of all variables, with fields
        "name", "offset" and "size".
        """
        node = self.remote_nodes[target_node_id]
        return [
            {
                "name": key,
                "offset": node.var_offset[key],
                "size": node.var_size[key]
            }
            for key in node.var_offset.keys()
        ]

    def get_var(self, target_node_id, name, index=0):
        """Get the value of a scalar variable from the local copy.
        """
        node = self.remote_nodes[target_node_id]
        with self.input_lock:
            return node.get_var(name, index)

    def get_var_array(self, target_node_id, name):
        """Get the value of an array variable from the local copy.
        """
        node = self.remote_nodes[target_node_id]
        try:
            with self.input_lock:
                return node.get_var_array(name)
        except KeyError:
            raise KeyError(name)

    def set_var(self, target_node_id, name, val, index=0):
        """Set the value of a scalar variable in the local copy and send it.
        """
        node = self.remote_nodes[target_node_id]
        with self.input_lock:
            node.set_var(name, val, index)
        self.set_variables(target_node_id, node.var_offset[name] + index, [val])

    def set_var_array(self, target_node_id, name, val):
        """Set the value of an array variable in the local copy and send it.
        """
        node = self.remote_nodes[target_node_id]
        with self.input_lock:
            node.set_var_array(name, val)
        self.set_variables(target_node_id, node.var_offset[name], val)

    def __getitem__(self, key):
        class Node:

            def __init__(self_node, connection, node_id):
                self_node.connection = connection
                self_node.node_id = node_id

            def __getitem__(self_node, name):
                try:
                    val = self_node.connection.get_var_array(self_node.node_id, name)
                    return val if len(val) != 1 else val[0]
                except KeyError:
                    raise KeyError(name)

            def __setitem__(self_node, name, val):
                try:
                    if isinstance(val, list):
                        self_node.connection.set_var_array(self_node.node_id, name, val)
                    else:
                        self_node.connection.set_var(self_node.node_id, name, val)
                except KeyError:
                    raise KeyError(name)

        return Node(self, key)

    def set_bytecode(self, target_node_id, bytecode, address=0):
        """Set the bytecode by sending one or more SET_BYTECODE messages.
        """
        size = len(bytecode)
        i = 0
        while i < size:
            size_chunk = min(size - i, 256)
            payload = Message.uint16array_to_bytes([
                target_node_id,
                address + i
            ] + bytecode[i:i + size_chunk])
            msg = Message(Message.ID_SET_BYTECODE, self.host_node_id, payload)
            self.send(msg)
            i += size_chunk

    def reset(self, target_node_id):
        """Reset the Thymio.
        """
        payload = Message.uint16array_to_bytes([
            target_node_id,
        ])
        msg = Message(Message.ID_RESET, self.host_node_id, payload)
        self.send(msg)

    def run(self, target_node_id):
        """Run the code on the Thymio.
        """
        payload = Message.uint16array_to_bytes([
            target_node_id,
        ])
        msg = Message(Message.ID_RUN, self.host_node_id, payload)
        self.send(msg)

    def get_device_info(self, target_node_id, info=None):
        """Request device info (all available by default).
        """
        if info is None:
            self.get_device_info(target_node_id,
                                 Message.DEVICE_INFO_NAME)
            self.get_device_info(target_node_id,
                                 Message.DEVICE_INFO_THYMIO2_RF_SETTINGS)
            self.get_device_info(target_node_id,
                                 Message.DEVICE_INFO_UUID)
        else:
            payload = Message.uint16array_to_bytes([
                target_node_id,
                info
            ])
            msg = Message(Message.ID_GET_DEVICE_INFO, self.host_node_id, payload)
            self.send(msg)
